/* -*- c-basic-offset: 4 indent-tabs-mode: nil -*-  vi:set ts=8 sts=4 sw=4: */

#define _USE_MATH_DEFINES
#include <cmath>

#include "devuvuzelator-vst.h"

#include <iostream>
#include <cstdio>

#include "params.h"

#define snprintf _snprintf
#define alloca _alloca

// VST params 0->1

void
Devuvuzelator::setParameter(VstInt32 index, float value)
{
    switch (index) {
    case 0: m_fundamental = 50 + 720 * value; break;
    case 1: m_bandwidth = 20 + 80 * value; break;
    case 2: m_harmonics = int(value * 6 + 0.5); break;
    case 3: m_reduction = 100 * value; break;
    }
}

float
Devuvuzelator::getParameter(VstInt32 index)
{
    switch (index) {
    case 0: return (m_fundamental - 50) / 720;
    case 1: return (m_bandwidth - 20) / 80;
    case 2: return (m_harmonics / 6.f);
    case 3: return m_reduction / 100;
    }
    return 0;
}

// NB! The max name length for VST parameter names, labels
// (i.e. units) and display values (i.e. string renderings of current
// value) is a rather amazing 8 bytes

void
Devuvuzelator::getParameterLabel(VstInt32 index, char *label)
{
    const char *units[NumParams] = {
        "Hz",
        "Hz",
        "",
        "%",
    };
    
    vst_strncpy(label, units[index], kVstMaxParamStrLen);
}

void
Devuvuzelator::getParameterDisplay(VstInt32 index, char *label)
{
    switch (index) {
    case 0: snprintf(label, kVstMaxParamStrLen, "%f", m_fundamental); break;
    case 1: snprintf(label, kVstMaxParamStrLen, "%f", m_bandwidth); break;
    case 2: snprintf(label, kVstMaxParamStrLen, "%d", m_harmonics); break;
    case 3: snprintf(label, kVstMaxParamStrLen, "%f", m_reduction); break;
    }
}

void
Devuvuzelator::getParameterName(VstInt32 index, char *label)
{
    const char *names[NumParams] = {
        "Pitch",
        "B/W",
        "Partials",
        "Reductn",
    };
    
    vst_strncpy(label, names[index], kVstMaxParamStrLen);
}

Devuvuzelator::Devuvuzelator(audioMasterCallback cb) :
    AudioEffect(cb, 0, NumParams),
    m_sampleRate(0),
    m_input(0),
    m_output(0),
    m_fftsize(FFTSIZE),
    m_winsize(WINSIZE),
    m_increment(m_winsize/2),
    m_filtersecs(FILTERSECS),
    m_fill(0),
    m_read(0)
{
    m_buffer = new float[m_winsize];
    m_outacc = new float[m_winsize * 2];
    m_window = new double[m_winsize];
    m_frame = new double[m_fftsize];
    m_spare = new double[m_fftsize];
    m_real = new double[m_fftsize];
    m_imag = new double[m_fftsize];
    m_medians = new MedianFilter<double> *[m_fftsize/2+1];

    for (int i = 0; i < m_winsize; ++i) {
        m_window[i] = 0.5 - 0.5 * cos(2 * M_PI * i / m_winsize);
    }
    for (int i = 0; i < m_fftsize/2+1; ++i) {
        m_medians[i] = 0;
    }

    m_fundamental = 230;
    m_bandwidth = 60;
    m_harmonics = 3;
    m_reduction = 30;
    
    setUniqueID('qmvz');
    setNumInputs(1);
    setNumOutputs(1);
    canProcessReplacing(true);
    canDoubleReplacing(false);

    reset();
}

Devuvuzelator::~Devuvuzelator()
{
    delete[] m_buffer;
    delete[] m_outacc;
    delete[] m_frame;
    delete[] m_spare;
    delete[] m_real;
    delete[] m_imag;
    delete[] m_window;
    for (int i = 0; i < m_fftsize/2+1; ++i) {
        delete m_medians[i];
    }
    delete[] m_medians;
}

void
Devuvuzelator::reset()
{
    for (int i = 0; i < m_winsize; ++i) {
        m_buffer[i] = 0.f;
    }
    for (int i = 0; i < m_winsize*2; ++i) {
        m_outacc[i] = 0.f;
    }
    m_fill = 0;
    m_read = 0;
    for (int i = 0; i < m_fftsize/2+1; ++i) {
        if (m_medians[i]) m_medians[i]->reset();
    }
}

void
Devuvuzelator::runImpl(unsigned long sampleCount)
{
    if (!m_input || !m_output) return;

    const int sc = sampleCount;

    float *output = m_output;
    if (m_input == m_output) {
        output = (float *)alloca(sampleCount * sizeof(float));
    }

    int oi = 0;
    for (int ii = 0; ii < sc; ++ii) {

        output[oi++] = m_outacc[m_read++];
        if (m_fill == m_winsize) {

            processFrame();

            for (int j = m_increment; j < m_winsize; ++j) {
                m_buffer[j - m_increment] = m_buffer[j];
            }

            for (int j = m_increment; j < m_winsize*2; ++j) {
                m_outacc[j - m_increment] = m_outacc[j];
            }

            for (int j = m_winsize*2 - m_increment; j < m_winsize*2; ++j) {
                m_outacc[j] = 0.f;
            }

            m_fill = m_fill - m_increment;
            m_read = m_read - m_increment;
        }
        m_buffer[m_fill] = m_input[ii];
        ++m_fill;
    }

    if (m_input == m_output) {
        for (int i = 0; i < sc; ++i) m_output[i] = output[i];
    }
}

void
Devuvuzelator::processFrame()
{
    for (int i = 0; i < m_fftsize; ++i) {
        m_frame[i] = 0.0;
    }

    int ix = m_fftsize - m_winsize/2;
    while (ix < 0) ix += m_fftsize;
    for (int i = 0; i < m_winsize; ++i) {
        m_frame[ix++] += m_buffer[i] * m_window[i];
        if (ix == m_fftsize) ix = 0;
    }

    fft(m_fftsize, false, m_frame, 0, m_real, m_imag);

    processSpectralFrame();

    for (int i = 0; i < m_fftsize/2-1; ++i) {
        m_real[m_fftsize-i-1] =  m_real[i+1];
        m_imag[m_fftsize-i-1] = -m_imag[i+1];
    }

    fft(m_fftsize, true, m_real, m_imag, m_frame, m_spare);

    ix = m_fftsize - m_winsize/2;
    while (ix < 0) ix += m_fftsize;
    for (int i = 0; i < m_winsize; ++i) {
        m_outacc[m_winsize + i] += m_frame[ix++];
        if (ix == m_fftsize) ix = 0;
    }
}

AudioEffect *createEffectInstance(audioMasterCallback audioMaster)
{
    return new Devuvuzelator(audioMaster);
}


